"""CLI handling."""

from pathlib import Path
from sys import exit
from typing import Optional

from arch_release_promotion import (
    argparse,
    config,
    files,
    gitlab,
    release,
    signature,
    torrent,
)


def promote_project_release(project: config.ProjectConfig, release_version: Optional[str] = None) -> None:
    """Promote a project's release.

    Parameters
    ----------
    project: config.ProjectConfig
        The ProjectConfig object that describes the project's configuration
    release_version: Optional[str]
        The optional release version to promote. If None is provided, interactive user input is required (defaults to
        None)
    """
    settings = config.Settings()  # type: ignore[call-arg]
    upstream = gitlab.Upstream(
        url=settings.GITLAB_URL,
        private_token=settings.PRIVATE_TOKEN,
        name=project.name,
    )
    upstream.auth()

    if not release_version:
        release_version = upstream.select_release()
        if not release_version:
            exit(1)

    artifact_temp_dir = files.create_temp_dir()
    promotion_temp_dir = files.create_temp_dir()

    artifact_zip = upstream.download_release(
        job_name=project.job_name,
        tag_name=release_version,
        temp_dir=artifact_temp_dir,
    )
    files.extract_zip_file_to_parent_dir(path=artifact_zip)

    for release_config in project.releases:
        artifact_output_path = artifact_temp_dir / project.output_dir
        artifact_release_path = artifact_output_path / Path(release_config.name)
        artifact_full_path = artifact_release_path / Path(f"{release_config.name}-{release_version}")
        promotion_base_path = promotion_temp_dir / Path("promotion")
        promotion_release_path = promotion_base_path / Path(release_config.name)
        promotion_full_path = promotion_release_path / Path(f"{release_config.name}-{release_version}")
        promotion_full_path.mkdir(parents=True)
        metrics_file = artifact_output_path / project.metrics_file

        signature.sign_files_in_dir(
            path=artifact_full_path,
            developer=settings.PACKAGER,
            gpgkey=settings.GPGKEY,
            file_extensions=release_config.extensions_to_sign,
        )

        metrics = files.read_metrics_file(
            path=metrics_file,
            version_metrics_names=release_config.version_metrics,
            size_metrics_names=release_config.size_metrics,
            amount_metrics_names=release_config.amount_metrics,
        )
        artifact_release = release.Release(
            name=release_config.name,
            version=release_version,
            files=files.files_in_dir(path=artifact_full_path),
            amount_metrics=metrics[0],
            size_metrics=metrics[1],
            version_metrics=metrics[2],
            torrent_file=torrent.create_torrent_file(
                path=artifact_full_path,
                webseeds=torrent.get_webseeds(
                    artifact_type=release_config.name,
                    mirrorlist_url=settings.MIRRORLIST_URL,
                    version=release_version,
                ),
                output=promotion_release_path / Path(f"{release_config.name}-{release_version}.torrent"),
            )
            if release_config.create_torrent
            else None,
            developer=settings.PACKAGER,
            pgp_public_key=settings.GPGKEY,
        )

        files.copy_signatures(source=artifact_full_path, destination=promotion_full_path)
        files.write_release_info_to_file(
            release=artifact_release,
            path=(promotion_release_path / Path(f"{release_config.name}-{release_version}.json")),
        )
        files.write_zip_file_to_parent_dir(path=promotion_base_path)

        upstream.promote_release(
            tag_name=release_version,
            file=str(promotion_temp_dir / Path("promotion.zip")),
        )

    files.remove_temp_dir(path=artifact_temp_dir)
    files.remove_temp_dir(path=promotion_temp_dir)


def main() -> None:
    """Run the main function."""
    args = argparse.ArgParseFactory.promote().parse_args()
    if args.project:
        project = config.Projects().get_project(name=args.project)  # type: ignore[call-arg]
        promote_project_release(project=project, release_version=args.release if args.release else None)
    else:
        for project in config.Projects().projects:  # type: ignore[call-arg]
            promote_project_release(project=project)


def arch_release_sync() -> None:
    """Synchronize releases.

    If the argument parser contains a specific project to synchronize, only synchronize that, else all configured
    projects.
    """
    args = argparse.ArgParseFactory.synchronize().parse_args()
    projects = config.Projects()  # type: ignore[call-arg]
    settings = config.Settings()  # type: ignore[call-arg]

    if args.project:
        files.ProjectFiles.sync(
            project_config=projects.get_project(name=args.project),
            settings=settings,
        )
    else:
        for project in projects.projects:
            files.ProjectFiles.sync(
                project_config=project,
                settings=settings,
            )



"""GitLab interaction."""

from pathlib import Path
from typing import List, Optional
from urllib import request

import gitlab


class Upstream(gitlab.Gitlab):
    """A class to interact with a gitlab instance.

    Upstream is derived from gitlab.Gitlab, but implements a few additional convenience methods such as
    self.select_release(), self.download_release() and self.promote_release().
    """

    def __init__(self, url: str, private_token: Optional[str], name: str):
        """Create an instance of Upstream.

        Parameters
        ----------
        url: str
            The URL of the gitlab instance to interact with
        private_token: Optional[str]
            An API token to use to interact with the gitlab instance
        name: str
            The name of the project to interact with
        """
        super().__init__(url=url, private_token=private_token)
        self.name = name
        if "/" in name:
            self.project_name = name.split("/")[-1]

    def select_release(self, max_releases: int = 3) -> Optional[str]:
        """Select one release from a project, that is not yet promoted.

        Parameters
        ----------
        max_releases: int
            How many releases to show at maximum

        Returns
        -------
        str
            The tag_name of the selected release
        """
        release_tags = self.get_releases(max_releases=max_releases, promoted=False)

        if release_tags:
            selection_string = "".join([f"{index}) {value}\n" for index, value in enumerate(release_tags)])
            selection = input(f"Select a release for {self.name}:\n{selection_string}")
            return release_tags[int(selection)]
        else:
            print("There are no releases to promote!")
            return None

    def get_releases(self, max_releases: int, promoted: bool = False) -> List[str]:
        """Get a list of releases of a project.

        Parameters
        ----------
        max_releases: int
            The maximum amount of releases to list
        promoted: bool
            Whether to only consider promoted releases (defaults to False)

        Returns
        -------
        List[str]
            A list of strings representing release tags of the project
        """
        project = self.projects.get(self.name)

        release_tags: List[str] = []

        for release in project.releases.list():
            if promoted:
                # only select releases that are promoted
                if (
                    any(link.name == "Promotion artifact" for link in release.links.list())
                    and len(release_tags) < max_releases
                ):
                    release_tags += [release.tag_name]
            else:
                # only select releases that are not yet promoted
                if (
                    not any(link.name == "Promotion artifact" for link in release.links.list())
                    and len(release_tags) < max_releases
                ):
                    release_tags += [release.tag_name]

        return release_tags

    def download_release(self, tag_name: str, temp_dir: Path, job_name: str) -> Path:
        """Download the build artifacts of a project's release as a compressed file.

        Parameters
        ----------
        tag_name: str
            The tag_name of the release to download
        temp_dir: Path
            The directory into which to download
        job_name: str
            The name of the job that provides the release artifacts

        Returns
        -------
        Path
            The file path of the downloaded compressed file
        """
        project = self.projects.get(self.name)
        artifact_zip = temp_dir / Path(f"{self.project_name}-{tag_name}.zip")
        try:
            print(f"Downloading build artifacts of release '{tag_name}' for '{self.name}'...")
            with open(artifact_zip, "wb") as download:
                project.artifacts(
                    ref_name=tag_name,
                    job=job_name,
                    streamed=True,
                    action=download.write,
                )
            print("Done!")
        except gitlab.exceptions.GitlabGetError:
            print(f"Skipping release {tag_name} as there are no artifacts to download")

        return artifact_zip

    def download_promotion_artifact(self, tag_name: str, temp_dir: Path) -> Path:
        """Download the promotion artifact of a project's release.

        Parameters
        ----------
        tag_name: str
            The tag_name of the release to download
        temp_dir: Path
            The directory into which to download

        Returns
        -------
        Path
            The file path of the downloaded file
        """
        project = self.projects.get(self.name)
        artifact_links: List[str] = []

        for release in project.releases.list():
            # only select releases that are promoted
            if release.tag_name == tag_name and any(link.name == "Promotion artifact" for link in release.links.list()):
                artifact_links += [link.url for link in release.links.list() if link.name == "Promotion artifact"]

        if not artifact_links:
            raise RuntimeError(
                f"There is no promotion artifact to download for the release '{tag_name}' of project '{project.name}'."
            )
        if len(artifact_links) > 1:
            raise RuntimeError(
                f"There is more than one promotion artifact to download for the release '{tag_name}' "
                f"of project '{project.name}'. Something is wrong!"
            )

        print(f"Downloading promotion artifact of release '{tag_name}' for '{self.name}'...")
        filename, headers = request.urlretrieve(  # ignore: B310
            artifact_links[0], filename=temp_dir / Path("promotion.zip")
        )
        print("Done!")

        return Path(filename)

    def promote_release(self, tag_name: str, file: str) -> None:
        """Upload a promotion file to the project and add it as a link to a release.

        Parameters
        ----------
        tag_name: str
            The tag_name of the project's release to attach a link to
        file: str
            The file path for a file to upload to the project
        """
        project = self.projects.get(self.name)
        print(f"Project '{self.name}': Uploading file {file}...")
        uploaded_file = project.upload(filename="promotion.zip", filepath=file)
        print(f"Project '{self.name}': Linking file {file} to release {tag_name}...")
        for release in project.releases.list():
            if release.tag_name == tag_name:
                release.links.create(
                    {"url": f"{self.url}/{self.name}/{uploaded_file['url']}", "name": "Promotion artifact"}
                )
